#pragma once

// Class:   CLVItemData 
// Purpose: Simple class that represents data contained in a list control item.
//          An instance of this class is created for every item in the list control.
// 
// Public Methods:
//      Cancel()        -- Cancels the retrieve and populate operations
//      Populate(...)   -- Populates a List control
//      Retrieve(...)   -- Retrieves data (fills a vector with data)
class CLVItemDataMgr
{
public:

    // Retrieval states
    enum { LVIDM_NOTRETRIEVED = 0,
            LVIDM_INPROGRESS = 1,
            LVIDM_RETRIEVED = 2 };
    
    CLVItemDataMgr()
        : m_hRetrieveThread( NULL )
            , m_hShutdownEvent( ::CreateEvent( NULL, TRUE, FALSE, NULL ) )
            , m_lMaxCount( 1000000 )
            , m_lRetrievalState( LVIDM_NOTRETRIEVED )
    {
    }

    ~CLVItemDataMgr()
    {
        ShutdownRetrieveThread();

        ::CloseHandle( m_hShutdownEvent );
        m_hShutdownEvent = NULL;
    }

// Operations (public methods consumed by CLVSelStateDlg)
public:
    // Method:  Cancel
    // Purpose: Called by CLVSelStateDlg::OnBnClickedCancel
    //          Shutsdown the secondary thread if running
    //          and clears the LVItem data list
    HRESULT Cancel()
    {
        ShutdownRetrieveThread();

        ::SetEvent( m_hShutdownEvent );
        ClearLVItemDataList();

        SetRetrievalState( LVIDM_NOTRETRIEVED );

        return S_OK;
    }

    // Method:  Populate
    // Purpose: Called within the Retrieve button to add items to
    // the list control.
    // NOTE: It can be argued that this type of data manager class shouldn't
    // include any UI interaction.  As such I could have cycled through the 
    // CLVItemDataList in the dialog itself and populated the list control.
    // However I chose not to because that would have mean I would have had
    // to publicly expose the CLVItemDataList member and force users to
    // lock the list in the dialog class. 
    HRESULT Populate( CListCtrl& ctlList )
    {
        HRESULT hr = S_OK;

        // Add a wait cursor
        CWaitCursor wait;

        long	lIndex = 0;
        LVITEM	lvitem = { 0 };
        int		iItem = 0, iActualItem = 0;

        
        lvitem.mask = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM;

        // Lock the List Control window so it doesn't look
        // like crap while filling
        ctlList.LockWindowUpdate();

        // Lock the list
        CAutoLockT< CLVItemDataList > lock( &m_LVItemDataList );

        // Cycle through the list and add each item into the list control
        for(CLVItemDataList::iterator it = m_LVItemDataList.begin();
                it < m_LVItemDataList.end();
                it++)
        {
            CLVItemData* pLVItemData = (*it);
	
		    lvitem.iItem        = iItem;
		    lvitem.iSubItem     = 0;
		    lvitem.lParam		= reinterpret_cast<LPARAM>(pLVItemData);

            // Tell the list control to become 'Virtual' and request display
            // data from the lvitem.lParam.
            // See CLVSelStateDlg::OnLvnGetdispinfoList for where this happens
            lvitem.iImage       = I_IMAGECALLBACK;
		    lvitem.pszText		= LPSTR_TEXTCALLBACK;
		    lvitem.cchTextMax	= MAX_LVITEMLEN;

            // Insert the item
		    ctlList.InsertItem(&lvitem);

            // Set the text for each column (all virtual callbacks)
		    ctlList.SetItemText(iItem, 0, LPSTR_TEXTCALLBACK);
		    ctlList.SetItemText(iItem, 1, LPSTR_TEXTCALLBACK);
		    ctlList.SetItemText(iItem, 2, LPSTR_TEXTCALLBACK);

		    iItem++;

            // Check if user cancelled operation 
            if( WAIT_OBJECT_0 == WaitForSingleObject( GetShutdownEvent(), 0 ) )
            {
                return hr;
            }

            // Pump messages to allow cancel button click
            // during control filling.  Populating the control
            // with thousands of records takes time, so we need
            // to process messages to handle the cancel button
            // and close messages
            if(!ProcessMessages( ))
            {
                return hr;
            }

            // Restore the wait cursor
            // (if you don't do this, the wait cursor may go
            // away after processing messages)
            wait.Restore();
	    }

        // Unlock the List control window
        ctlList.UnlockWindowUpdate();

        return hr;
    }

    // Method:  Populate
    // Purpose: Called within CLVSelStateDlg::OnBnClickedRetrieve()
    // Spawns a secondary thread which adds items to the LVItem data list
    // and then waits for the thread to finish
    HRESULT Retrieve( const LONG lMaxCount )
    {
        HRESULT hr = S_OK;
        UINT uRetrieveTID = 0;

        // Stop and clean up if there is an existing thread
        ShutdownRetrieveThread();
        ClearLVItemDataList();
        ::ResetEvent( m_hShutdownEvent );

        SetMaxCount( lMaxCount );
        SetRetrievalState( LVIDM_INPROGRESS );


        if( NULL == ( m_hRetrieveThread
                        = (HANDLE) _beginthreadex( NULL,
                                                    0,
                                                    RetrieveThread,
                                                    this,
                                                    0,
                                                    &uRetrieveTID )))
        {
            return HRESULT_FROM_WIN32( ::GetLastError() );
        }

        // Process messages and wait for the thread to finish
        return MsgWaitForThread( m_hRetrieveThread );
    }

// Implementation
private:

    // Method:  ClearLVItemDataList
    // Purpose: Deletes any CLVItemData objects and clears the list
    void    ClearLVItemDataList()
    {
        CAutoLockT< CLVItemDataList > lock( &m_LVItemDataList );

        // Delete CLVItemDataList pointers
	    for(CLVItemDataList::iterator it = m_LVItemDataList.begin();
                it < m_LVItemDataList.end();
                it++)
	    {
		    delete (*it);
	    }

        m_LVItemDataList.clear();
    }

    // Method:  InsertItem
    // Purpose: Creates a new CLVItemData object and inserts it into the list
    HRESULT InsertItem( const UINT uIndex )
    {   
        HRESULT hr = S_OK;

        CString sColumnOne = _T(""), sColumnTwo = _T("Col2"), sColumnThree = _T("Col3");

        sColumnOne.Format( _T("Col1 : %d"), uIndex );

        CLVItemData* pLVItemData = new CLVItemData( sColumnOne, sColumnTwo, sColumnThree );

        if( NULL == pLVItemData )
        {
            return E_OUTOFMEMORY;
        }

        try
        {
            CAutoLockT< CLVItemDataList > lock( &m_LVItemDataList );
            m_LVItemDataList.push_back( pLVItemData );
        }
        catch(std::bad_alloc)
        {
            hr = E_OUTOFMEMORY;
        }

        return hr;
    }

    // Method:  GetMaxCount
    // Purpose: Returns the MaxCount (thread safe)
    const LONG GetMaxCount()
    {
        return ::InterlockedExchange( &m_lMaxCount, m_lMaxCount );
    }

    // Method:  GetRetrievalState
    // Purpose: Returns the current retrieval state (thread safe)
    const UINT GetRetrievalState( )
    {
        return ::InterlockedExchange( &m_lRetrievalState, m_lRetrievalState );
    }

    // Method:  GetShutdownEvent
    // Purpose: Returns the shutdown event handle
    HANDLE& GetShutdownEvent()
    {
        return m_hShutdownEvent;
    }

    // Method:  MsgWaitForThread
    // Purpose: Processes messages and waits for the specify thread
    HRESULT MsgWaitForThread( HANDLE hThread )
    {
        HRESULT hr = S_OK;
        BOOL bContinue = TRUE;
        DWORD dwSignalled = 0;

        CWaitCursor wait;

	    // Process main thread messages
		MSG msg = { 0 };

	    while ( bContinue )
	    {
		    dwSignalled = MsgWaitForMultipleObjects( 1,
												    &hThread,
												    FALSE,
												    INFINITE,
												    QS_ALLINPUT);

		    switch ( dwSignalled )
		    {
		    case WAIT_OBJECT_0:
                // Thread exited, so bail
			    bContinue = FALSE;
			    break;
		    case WAIT_OBJECT_0 + 1:
                // Process windows messages
                bContinue = ProcessMessages();
                wait.Restore(); // Restore the wait cursor
			    break;
		    default:
			    return HRESULT_FROM_WIN32( ::GetLastError() );
		    }
        }
        return hr;
	}		

    // Method:  ProcessMessages
    // Purpose: Processes windows messages
    BOOL ProcessMessages( )
    {
        MSG msg = { 0 };

        while( ::PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) )
		{
			if( WM_QUIT == msg.message )
			{
				// Quit message received
				SetEvent( m_hShutdownEvent );
				return FALSE;
			}
			else
			{
				// Translate and dispatch the message
				TranslateMessage( &msg );
				DispatchMessage( &msg );
			}
		}
        return TRUE;
    }

    // Method:  RetrieveThread
    // Purpose: Secondary Thread which adds Max Count items to the
    // LVItem Data list
    static UINT WINAPI RetrieveThread( LPVOID lpContext )
    {
        CLVItemDataMgr* pLVItemDataMgr = static_cast< CLVItemDataMgr* >( lpContext );

        for( LONG lIndex = 0; lIndex < pLVItemDataMgr->GetMaxCount(); lIndex++ )
        {
            if( WAIT_OBJECT_0 == WaitForSingleObject( pLVItemDataMgr->GetShutdownEvent(), 0 ) )
            {
                break;
            }

            pLVItemDataMgr->InsertItem( lIndex );
        }

        pLVItemDataMgr->SetRetrievalState( LVIDM_RETRIEVED );

        return 0;
    }

    // Method:  SetMaxCount
    // Purpose: Sets the MaxCount (thread safe)
    void SetMaxCount( const LONG lMaxCount )
    {
        ::InterlockedExchange( &m_lMaxCount, lMaxCount );
    }

    // Method:  SetRetrievalState
    // Purpose: Sets the current retrieval state (thread safe)
    void SetRetrievalState( const LONG lState )
    {
        ::InterlockedExchange( &m_lRetrievalState, lState );
    }

    // Method:  ShutdownRetrieveThread
    // Purpose: Closes the Sets the MaxCount (thread safe)
    void ShutdownRetrieveThread()
    {
	    // Close the Retrieve thread
	    if( NULL != m_hRetrieveThread )
	    {
		    ::SetEvent( m_hShutdownEvent );
		    
            if ( WAIT_TIMEOUT == WaitForSingleObject( m_hRetrieveThread, 1000 ) ) 
            {
                TRACE( _T("Using TerminateThread on Retrieve Thread\n") );
                TerminateThread( m_hRetrieveThread, -1000 );
            }

		    ::CloseHandle( m_hRetrieveThread );
		    m_hRetrieveThread = NULL;
	    }
    }

// Attributes
private:
    HANDLE          m_hShutdownEvent;
    HANDLE          m_hRetrieveThread;
    CLVItemDataList m_LVItemDataList;
    LONG            m_lMaxCount;
    LONG            m_lRetrievalState;

};


